/*
# $BEGIN_KEYHOLE_LICENSE$
# 
# This file is part of the Keyhole project, a lightweight library for
# the Arduino IDE, for interpreting commands and communicating variable
# values via a JSON- and Python-compatible text interface. 
# 
# Copyright (c) 2023- Jeremy Hill
# 
# Behold the 3-Clause BSD License under which it is hereby released:
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived
#    from this software without specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# 
# $END_KEYHOLE_LICENSE$

The following example allows two global variables to be queried and/or
set over a serial port or other Stream::

    #include "Keyhole.h"
  
    double foo = 0.0;
    String bar = "look at these escaped characters: \" \x08 \\ \t\r\n";
  
    void setup()
    {
      Serial.begin(9600);
    }
  
    void loop()
    {
      // main business of your sketch goes here
      // ...
      
      KEYHOLE keyhole(Serial);
      if( keyhole.begin() )
      {
        if(keyhole.variable("foo", foo)) Serial.println("assigned to foo");
        if(keyhole.variable("bar", bar)) Serial.println("assigned to bar");
        if(keyhole.command("Marco!")) Serial.println("Polo!");
        keyhole.end();
      }
    }
    
An optional third argument to `variable()` can be `VARIABLE_READ_ONLY`,
`VARIABLE_VERBOSE`, or the default value `VARIABLE_SILENT`.

In this example, if you send the command `foo` over the serial port,
terminated either by a semicolon or a newline (the latter of which
gets added automatically by the Arduino IDE's Serial Monitor) the
reply will be a JSON string associating the key `foo` with the
current value of the corresponding global variable `foo`. 

Provided the variable is not designated as `VARIABLE_READ_ONLY`, you can
also set its value, e.g. by sending the command `foo = 2`.  (If you have
designated the variable as `VARIABLE_VERBOSE` then this will also trigger
a JSON output, as if you had queried `foo` immediately after setting it.)

You can send the simple command `?` to receive a JSON output containing all
the variables that are accessible in this manner.

You can also have the full report delivered automatically on a repeating
schedule by setting the `.autoSeconds` member greater than zero; you can
even allow this parameter itself to be controlled via the keyhole, by
exposing the `autoSeconds` member itself as a variable::

      KEYHOLE keyhole(Serial);
      if(keyhole.begin())
      {
        if(keyhole.variable("foo", foo)) Serial.println("assigned to foo");
        if(keyhole.variable("bar", bar)) Serial.println("assigned to bar");
        keyhole.variable("report_period_sec", keyhole.autoSeconds);
        keyhole.end();
      }

Under the hood: `keyhole` is an instance of the class `Keyhole`. Such
instances must either be global variables, or declared `static`. The
`KEYHOLE` macro can be used inside or outside of `loop()` - it simply
declares a `static Keyhole` instance (while hiding the possibly-not-so-
user-friendly keyword `static` from users who are not familiar with it).
  
*/

#ifndef   __Keyhole_H__
#define   __Keyhole_H__

#include "Arduino.h" // for Stream and String

// Debugging macros:
#define DBSTREAM Serial
#define REPORT( X )  { DBSTREAM.print( "{\"" #X "\" : " ); DBSTREAM.print( X );                         DBSTREAM.println( "}" ); DBSTREAM.flush(); }
#define REPORTS( X ) { DBSTREAM.print( "{\"" #X "\" : " ); Keyhole( DBSTREAM ).printStringLiteral( X ); DBSTREAM.println( "}" ); DBSTREAM.flush(); }

typedef enum
{
	VARIABLE_READ_ONLY = 0,
	VARIABLE_SILENT    = 1,
	VARIABLE_VERBOSE   = 2
} KeyholeWriteMode;

#define KEYHOLE       static Keyhole
class Keyhole
{
	public:
		Keyhole( Stream & stream=Serial, float autoSeconds=0.0, bool plotterMode=false );
		~Keyhole();
	
		bool begin( void );
		bool begin( unsigned long microsecondTimestamp );
	
		bool command( const char * cmd );
	
		// A char may be considered signed or unsigned, depending on architecture.
		// So, in case it is unsigned, we need an explicit signed-8-bit polymorph:
		bool variable( const char * key, int8_t &         referenceToVariable,  KeyholeWriteMode mode=VARIABLE_SILENT ); 
		// The rest of the stdint types should take care of themselves via the usual builtin types:
		bool variable( const char * key, bool &           referenceToVariable,  KeyholeWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * key, char &           referenceToVariable,  KeyholeWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * key, unsigned char &  referenceToVariable,  KeyholeWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * key, int &            referenceToVariable,  KeyholeWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * key, unsigned int &   referenceToVariable,  KeyholeWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * key, short &          referenceToVariable,  KeyholeWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * key, unsigned short & referenceToVariable,  KeyholeWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * key, long &           referenceToVariable,  KeyholeWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * key, unsigned long &  referenceToVariable,  KeyholeWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * key, float &          referenceToVariable,  KeyholeWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * key, double &         referenceToVariable,  KeyholeWriteMode mode=VARIABLE_SILENT );
		bool variable( const char * key, String &         referenceToVariable,  KeyholeWriteMode mode=VARIABLE_SILENT );
		
#		define variableAssigned variable // so you can express it like this if the semantics appeal to you more:
		                                 //     if( keyhole.variableAssigned("foo", foo) ) doWhatever(foo);
	
		bool end( void );
	
		unsigned long elapsedMicros( void ); // only gives meaningful results if you passed micros() to .begin(), or if .autoSeconds is set

		Stream &      stream;
		float         autoSeconds; // set this >0.0 to receive periodic automatic output
		int           plotterMode; // set this to true to make the output format compatible with the Serial Plotter in the Arduino IDE
		
	private: // nothing to see here
		unsigned long mBeginMicros;
		String        mFullCommand;
		int           mListAllVariables;
		String        mPartialCommand;
		bool          mBackslash;
		int           mHexEscape;
		char          mHexValue;
		char          mQuote;
		unsigned long mTimestampOfLastAutoReport;
		char          mBad;

		const char *  _parseVariableCommand( const char * key, unsigned int & commandLength );

	public:
		// Some general-purpose helper methods (they're instance methods only because they access `this->stream`)
		void          printLiteral( double f,         char withQuotes='"' );
		void          printLiteral( const String & s, char withQuotes='"' );
		void          printLiteral( char c,           char withQuotes=(char)-1 );
		// * withQuotes=0 means print characters (or escape codes) unquoted, and don't escape any quotes.
		// * withQuotes='\'' or withQuotes='"' means print the content wrapped in the specified type of quote, and
		//   escape that quote character if it occurs in the content. (For floating-point, quotes are only used if the
		//   value is an inf or a nan, for reasons of JSON legality and ast.literal_eval() compatibility.)
		// * withQuotes=-1 has a special meaning for chars which causes the char to be printed as a numeric int8_t or
		//   uint8_t value (depending on architecture). This is the default for reasons discussed in the comments of the
		//   method.
	
		////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		// And finally some general-purpose static helper functions (but let's keep them inside the Keyhole:: namespace):
	
		// These are needed because some boards' standard libraries don't have a String(ptr,length) constructor 
		// and/or may mess up copying `string2=string1;` if `string1` contains '\0'
		static void          assignString( String & dst, const String & src, bool trim=false, bool lowercase=false );
		static void          assignString( String & dst, const char *ptr, unsigned int length, bool trim=false, bool lowercase=false );
		
		// These work the same as the standard library functions `strtoul`, `strtol` and `strtod` except that they
		// handle 0x and 0b prefixes and infer the base (16, 2 or 10):
		static unsigned long strToUnsignedInteger( const char * start, char ** endptr );
		static long          strToSignedInteger(   const char * start, char ** endptr );
		static double        strToDouble(          const char * start, char ** endptr );
		
		// This one is good at the start or end of `setup()`, to say hello and reassure the user that things are working:
#ifdef LED_BUILTIN
		static void          flicker( int ledPin=LED_BUILTIN, unsigned long millisOn=50, unsigned long millisOff=50, unsigned long millisTotal=1000 );
#else
		static void          flicker( int ledPin,             unsigned long millisOn=50, unsigned long millisOff=50, unsigned long millisTotal=1000 );
#endif
};

#endif // __Keyhole_H__
